/*
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This file is public domain.
 */

#include <glib.h>
#include <stdio.h>
#include <string.h>                     /* strcmp() */
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "debug.h"

#ifdef DEBUG
typedef enum {T_WIDGETDB,T_WIDGET,T_PROPERTY} NodeType;
# define TYPE_CHECK(n,t,ts) if(**(NodeType**)n!=t) g_error("node is %i not <" ts "> in %s(): %i",**(int**)n,__func__,__LINE__)
# define TYPE_CHECK_WIDGETDB(n) TYPE_CHECK(n,T_WIDGETDB,"widgetdb")
# define TYPE_CHECK_WIDGET(n)   TYPE_CHECK(n,T_WIDGET,"widget")
# define TYPE_CHECK_PROPERTY(n) TYPE_CHECK(n,T_PROPERTY,"property")
#else
# define TYPE_CHECK_WIDGETDB(n)
# define TYPE_CHECK_WIDGET(n)
# define TYPE_CHECK_PROPERTY(n)
#endif

typedef struct{
#ifdef DEBUG
    NodeType node_type;
#endif
    xmlDoc* main_doc;
    xmlNode* main_root;
    GHashTable* widget_hash;
}WidgetdbInfo;

typedef struct{
#ifdef DEBUG
    NodeType node_type;
#endif
    char* class;
    char* parent;
    char* constructor;
    char* linker;
    GHashTable* property_hash;
}WidgetInfo;

typedef struct{
#ifdef DEBUG
    NodeType node_type;
#endif
    char* name;
    char* type;
    char* default_value;
    char* setter;
}PropertyInfo;



WidgetdbInfo state;



void property_cache(xmlNode* property){
    PropertyInfo* info;
    xmlNode* iter;
    xmlBuffer* buffer;
    char** s;
    if(!property->_private){
        property->_private = info = g_new0(PropertyInfo,1);
#ifdef DEBUG
        info->node_type = T_PROPERTY;
#endif
    }else{
        TYPE_CHECK_PROPERTY(property);
        info = property->_private;
        if(info->name && info->type && info->default_value && info->setter)
            return;
    }
    TYPE_CHECK_PROPERTY(property);
    for(iter=property->children;iter;iter=iter->next){
        if(iter->type != XML_ELEMENT_NODE)
            continue;
        if(!strcmp((char*)iter->name,"name")){
            s = &info->name;
        }else if(!strcmp((char*)iter->name,"type")){
            s = &info->type;
        }else if(!strcmp((char*)iter->name,"default_value")){
            s = &info->default_value;
        }else if(!strcmp((char*)iter->name,"setter")){
            s = &info->setter;
        }else{
            continue;
        }
        if(*s)
            continue;
        buffer = xmlBufferCreateSize(64);
        if(xmlNodeBufGetContent(buffer,iter)>=0){
            *s = g_strdup((char*)xmlBufferContent(buffer));
            DEBUG_p("setting %s = \"%s\"",iter->name,*s);
        }
        xmlBufferFree(buffer);
    }
}

static char* property_get_name(xmlNode* property){
    PropertyInfo* info;
    info = property->_private;
    if(!info || !info->name)
        property_cache(property);
    TYPE_CHECK_PROPERTY(property);
    info = property->_private;
    return info->name;
}

static char* property_get_type(xmlNode* property){
    PropertyInfo* info;
    info = property->_private;
    if(!info || !info->type)
        property_cache(property);
    TYPE_CHECK_PROPERTY(property);
    info = property->_private;
    return info->type;
}



static void widget_cache(xmlNode* widget){
    WidgetInfo* info;
    xmlNode* iter;
    xmlBuffer* buffer;
    char** s;
    if(!widget->_private){
        DEBUG_p("this node has no info yet");
        widget->_private = info = g_new0(WidgetInfo,1);
#ifdef DEBUG
        info->node_type = T_WIDGET;
#endif
        info->property_hash = g_hash_table_new(g_str_hash,g_str_equal);
    }else{
        TYPE_CHECK_WIDGET(widget);
        info = widget->_private;
        if(!info->property_hash)
            g_warning("property hash is somehow NULL");
        if(info->class && info->parent && info->constructor && info->linker)
            return;
    }
    for(iter=widget->children;iter;iter=iter->next){
        if(iter->type != XML_ELEMENT_NODE)
            continue;
        if(!strcmp((char*)iter->name,"class")){
            s = &info->class;
        }else if(!strcmp((char*)iter->name,"parent")){
            s = &info->parent;
        }else if(!strcmp((char*)iter->name,"constructor")){
            s = &info->constructor;
        }else if(!strcmp((char*)iter->name,"linker")){
            s = &info->linker;
        }else{
            continue;
        }
        if(*s)
            continue;
        buffer = xmlBufferCreateSize(64);
        if(xmlNodeBufGetContent(buffer,iter)>=0){
            *s = g_strdup((char*)xmlBufferContent(buffer));
            DEBUG_p("setting %s = \"%s\"",iter->name,*s);
        }
        xmlBufferFree(buffer);
    }
}

static char* widget_get_class(xmlNode* widget){
    WidgetInfo* info;
    info = widget->_private;
    if(!info || !info->class)
        widget_cache(widget);
    info = widget->_private;
    return info->class;
}

static char* widget_get_parent(xmlNode* widget){
    WidgetInfo* info;
    info = widget->_private;
    if(!info || !info->parent)
        widget_cache(widget);
    info = widget->_private;
    return info->parent;
}

static char* widget_get_constructor(xmlNode* widget){
    WidgetInfo* info;
    info = widget->_private;
    if(!info || !info->constructor)
        widget_cache(widget);
    info = widget->_private;
    return info->constructor;
}



static void merge_property_nodes(xmlNode* dst,xmlNode* src){
    xmlNode* node;
    TYPE_CHECK_PROPERTY(dst);
    TYPE_CHECK_PROPERTY(src);
    /* the following items should be moved from b to a if they exist in b
       but a: <type> <default> <setter> */
    if(property_get_type(src)){
        if(!property_get_type(dst)){
            node = xmlNewNode(NULL,(xmlChar*)"type");
            xmlNodeSetContent(node,(xmlChar*)property_get_type(src));
            DEBUG_p("adding %s to %s",node->name,widget_get_class(dst));
            xmlAddChild(dst,node);
            xmlAddChild(dst,xmlNewText((xmlChar*)"\n"));
        }
    }
    xmlUnlinkNode(dst);
}

static gboolean merge_property_tree(xmlNode* dst_widget,xmlNode* src_widget){
    WidgetInfo* info;
    xmlNode* src;
    xmlNode* dst;
    xmlNode* tmp;
    DEBUG_p("begin");
    info = dst_widget->_private;
    DEBUG_p("info = %xh",(int)info);
    for(src=src_widget->children;src;src=tmp){
        tmp = src->next;
        switch(src->type){
        case XML_ELEMENT_NODE:
            if(strcmp((char*)src->name,"property")){
                DEBUG_p("skipping entry <%s>",src->name);
                continue;
            }
            DEBUG_p("property \"%s\" is of type \"%s\"",property_get_name(src),property_get_type(src));
            /* try to find this property in the hash in order to merge it */
            DEBUG_p("property hash = %xh",(int)info->property_hash);
            dst = g_hash_table_lookup(info->property_hash,property_get_name(src));
            DEBUG_p("dst property = %xh",(int)dst);
            if(dst){
                /* the property was found -> merge */
                DEBUG_p("property \"%s\" was found in the hash",property_get_name(src));
                merge_property_nodes(dst,src);
            }else{
                /* it is unique -> add it to the hash */
                DEBUG_p("property \"%s\" is unique",property_get_name(src));
                if(dst_widget != src_widget){
                    DEBUG_p("adopt property from a different widget node");
#if 1
# if 1
                    src = xmlCopyNode(src,1);
                    xmlAddChild(dst_widget,src);
                    xmlAddChild(dst,xmlNewText((xmlChar*)"\n"));
# else
                    xmlUnlinkNode(widget_b);
                    xmlDOMWrapAdoptNode(NULL,doc,widget_b,state.main_doc,state.main_root,0);
# endif
#endif
                }
                g_hash_table_insert(info->property_hash,property_get_name(src),src);
            }
            break;
        case XML_TEXT_NODE:
        default:
            break;
        }
    }
    DEBUG_p("end");
    return TRUE;
}



static void merge_widget_nodes(xmlNode* dst,xmlNode* src){
    xmlNode* node;
    TYPE_CHECK_WIDGET(dst);
    TYPE_CHECK_WIDGET(src);
    /* the following items should be moved from b to a if they exist in b
       but a: <parent> <constructor> <linker> */
    if(widget_get_constructor(src)){
        if(!widget_get_constructor(dst)){
            node = xmlNewNode(NULL,(xmlChar*)"constructor");
            xmlNodeSetContent(node,(xmlChar*)widget_get_constructor(src));
            DEBUG_p("adding %s to %s",node->name,widget_get_class(dst));
            xmlAddChild(dst,node);
            xmlAddChild(dst,xmlNewText((xmlChar*)"\n"));
        }
    }

    DEBUG_p("merging properties");
    /* make sure the widget has a property hash */
    widget_cache(dst);
    /* the following elements should be moved or merged if they exist in b:
       <properies> */
    /* first merge all properties if dst widget (if there are dupes) */
    merge_property_tree(dst,dst);
    /* then merge in the properties from src */
    merge_property_tree(dst,src);
}

gboolean merge_widget_tree(xmlDoc* doc,xmlNode* root){
    xmlNode* src;
    xmlNode* dst;
    xmlNode* tmp;

    if(strcmp((char*)root->name,"widgetdb")){
        g_printerr("Root element is not <widgetdb>.\n");
        return FALSE;
    }
    for(src=root->children;src;src=tmp){
        tmp = src->next;
        switch(src->type){
        case XML_ELEMENT_NODE:
            if(strcmp((char*)src->name,"widget")){
                DEBUG_p("skipping entry <%s>",src->name);
                continue;
            }
            widget_cache(src);
            DEBUG_p("src.info = %xh",(int)src->_private);
            TYPE_CHECK_WIDGET(src);
            /* try to find this widget in the hash in order to merge it */
            DEBUG_p("looking up %s",widget_get_class(src));
            dst = g_hash_table_lookup(state.widget_hash,widget_get_class(src));
            /* DEBUG_p("dst = %xh",(int)dst); */
            if(dst){
                DEBUG_p("looked up %s (%s)",dst->name,widget_get_class(dst));
                TYPE_CHECK_WIDGET(dst);
                /* the widget was found -> merge */
                DEBUG_p("widget \"%s\" was found in the hash",widget_get_class(src));
                merge_widget_nodes(dst,src);
            }else{
                /* it is unique -> add it to the hash */
                DEBUG_p("widget \"%s\" is unique",widget_get_class(src));
                TYPE_CHECK_WIDGET(src);
                if(doc!=state.main_doc){
#if 1
                    src = xmlCopyNode(src,1);
                    xmlAddChild(state.main_root,src);
                    xmlAddChild(state.main_root,xmlNewText((xmlChar*)"\n"));
#else
                    xmlUnlinkNode(src);
                    xmlDOMWrapAdoptNode(NULL,doc,src,state.main_doc,state.main_root,0);
#endif
                }
                g_hash_table_insert(state.widget_hash,widget_get_class(src),src);
            }
            break;
        case XML_TEXT_NODE:
        default:
            break;
        }
    }
    return TRUE;
}



int main(int argc,char** argv){
    int i;

    if(argc<2){
        g_printerr("Usage: %s <filename.widgetdb>\n",argv[0]);
        return(1);
    }

    /*
     * this initialize the library and check potential ABI mismatches
     * between the version it was compiled for and the actual shared
     * library used.
     */
    LIBXML_TEST_VERSION

    /* Parse the file and get the DOM */
    state.main_doc = xmlReadFile(argv[1], NULL, 0);
    if(!state.main_doc){
        g_printerr("Could not parse file %s.\n",argv[1]);
        return 1;
    }

    /* Get the root element node */
    state.main_root = xmlDocGetRootElement(state.main_doc);

    /* Create the widget class hash to quickly compare widgets */
    state.widget_hash = g_hash_table_new(g_str_hash,g_str_equal);

    /* Merge the main document */
    if(!merge_widget_tree(state.main_doc,state.main_root))
        return 1;

    /* Merge in other documents */
    for(i=2;i<argc;++i){
        xmlDoc* doc;
        xmlNode* root;
        DEBUG_p("reading %s",argv[i]);
        doc = xmlReadFile(argv[i],NULL,0);
        if(!doc){
            g_printerr("Could not parse file %s.\n",argv[i]);
            continue;
        }
        root = xmlDocGetRootElement(doc);
        merge_widget_tree(doc,root);
        xmlFreeDoc(doc);
    }

    /* Dumping document to stdio or file */
    DEBUG_p("writing output");
    xmlSaveFormatFileEnc("-",state.main_doc,"UTF-8",1);

    /* Free the document */
    xmlFreeDoc(state.main_doc);

    /* Free the global variables that may have been allocated by the
       parser. */
    xmlCleanupParser();

    return 0;
}
