/*
 * bin2c - convert any data to C source
 *
 * Copyright (C) 2009 Magnus Bergman
 *
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../autoconf.h"

#include <fcntl.h>
#include <glib.h>
#include <stdlib.h>                     /* exit() */
#include <unistd.h>

#ifdef USE_COMPRESSION
# include "compression.h"
#endif
#include "debug.h"

#define BUFFER 4096
#define ROWS 12

#ifndef ABS
# define ABS(a) ((a)<0?-(a):(a))
#endif

#ifdef USE_COMPRESSION
gboolean opt_compress;
#endif
gboolean opt_header;
gboolean opt_version;

static GOptionEntry entries[] = {
#ifdef USE_COMPRESSION
  {"compress",        'z', 0,G_OPTION_ARG_NONE,&opt_compress,"Compress the file",NULL},
#endif
  {"header",          'h', 0,G_OPTION_ARG_NONE,&opt_header,"Output header instead of source","DIR"},
  {"version",         'V', 0,G_OPTION_ARG_NONE,&opt_version,"Display version information",NULL},
  {NULL}
};

static void print_usage(char* prg){
    g_printerr("Usage: %s [OPTIONS] <file> <name>\n",prg);
    exit(1);
}

static int nr;
static void write_output(void* data,int bytes,void* userdata){
    unsigned char* s;
    for(s=data;bytes>0;++s,--bytes,++nr){
        if(nr>0){
            printf(",");
            if(nr%ROWS<=0)
                printf("\n");
        }
        if(nr%ROWS<=0)
            printf("   ");
        printf(" 0x%02X",*s);
    }
}
#ifdef USE_COMPRESSION
static void count_output(void* data,int bytes,void* userdata){
    nr+=bytes;
}
#endif

int main(int argc,char** argv){
    int fd;
    int size;
    int compressed_size;

    {
        GOptionContext* context;
        context = g_option_context_new("<file> <name>");
        g_option_context_add_main_entries(context,entries,NULL);
        if(!g_option_context_parse(context,&argc,&argv,NULL))
            print_usage(argv[0]);
    }

    if(argc<3)
        print_usage(argv[0]);

    fd = open(argv[1],O_RDONLY);
    size = lseek(fd,0,SEEK_END);
    lseek(fd,0,SEEK_SET);

    if(opt_header){
        g_print("#ifndef __%s_h__\n#define __%s_h__\n\n",argv[2],argv[2]);
#ifdef USE_COMPRESSION
        if(opt_compress){
            unsigned char buffer[BUFFER];
            int bytes;
            CompressionContext* context;
            g_print("#define %s_is_compressed 1\n",argv[2]);
            /* compress the data just to get the size, maybe not the optimal solution */
            context = compress_init(count_output,(void*)fd);
            for(;;){
                bytes = read(fd,buffer,BUFFER);
                if(bytes<=0)
                    break;
                compress_feed(context,buffer,bytes);
            }
            compress_done(context);
            compressed_size = nr;
        }else{
#else
        {
#endif
            /* there is no compression */
            compressed_size = size;
        }
        /* this is the compressed size of the data (if compressed) */
        g_print("#define %s_size %i\n",argv[2],compressed_size);
        g_print("#define %s_uncompressed_size %i\n",argv[2],size);
        g_print("unsigned char %s[%i];\n",argv[2],compressed_size);
        g_print("\n#endif /* __%s_h__ */\n",argv[2]);
    }else{
        /* if the data is compressed we don't know the size, but that's ok */
        g_print("unsigned char %s[] = {\n",argv[2]);

#ifdef USE_COMPRESSION
        if(opt_compress){
            unsigned char buffer[BUFFER];
            int bytes;
            CompressionContext* context;
            context = compress_init(write_output,(void*)fd);
            for(;;){
                bytes = read(fd,buffer,BUFFER);
                if(bytes<=0)
                    break;
                compress_feed(context,buffer,bytes);
            }
            compress_done(context);
        }else{
#else
        {
#endif
            unsigned char buffer[BUFFER];
            int bytes;
            for(;;){
                bytes = read(fd,buffer,BUFFER);
                if(bytes<=0)
                    break;
                write_output(buffer,bytes,(void*)fd);
            }
        }
        printf("\n};\n");
    }
    close(fd);
    return 0;
}
