/*
 * simple wrapper around zlib
 *
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This file is public domain.
 */

#include <stdlib.h>                     /* malloc() free() */
#include <zlib.h>

#include "compression.h"
#include "debug.h"

#define BUFFER 256

struct _CompressionContext{
    z_stream z;
    unsigned char buffer[BUFFER];
    CompressionCallback callback;
    void* userdata;
};

typedef struct{
    int(*init)(z_stream*);
    int(*feed)(z_stream*,int);
    int(*done)(z_stream*);
}StreamInterface;

/* This is needed to set the extra parameter */
static int deflateInit_wrapper(z_stream* z){
    return deflateInit(z,Z_DEFAULT_COMPRESSION);
}

/* This is needed because it's a macro in zlib */
static int inflateInit_wrapper(z_stream* z){
    return inflateInit(z);
}



static CompressionContext* stream_init(StreamInterface* sif,CompressionCallback callback,void* userdata){
    CompressionContext* context;
    context = (CompressionContext*)malloc(sizeof(CompressionContext));
    context->z.zalloc = Z_NULL;
    context->z.zfree = Z_NULL;
    if(sif->init(&context->z) != Z_OK){
        DEBUG_p("error calling ::init()");
        free(context);
        return NULL;
    }
    context->z.next_out = context->buffer;
    context->z.avail_out = BUFFER;
    context->callback = callback;
    context->userdata = userdata;
    return context;
}

static void stream_feed(StreamInterface* sif,CompressionContext* context,void* data,int bytes){
    int status;
    int count;

    DEBUG_p("begin");
    context->z.next_in = data;
    context->z.avail_in = bytes;

    while(context->z.avail_in){
        DEBUG_p("before deflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
        status = sif->feed(&context->z,Z_NO_FLUSH);
        DEBUG_p("after deflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
        count = BUFFER - context->z.avail_out;
        DEBUG_p("%i available to write",count);
        if(!count)
            continue;
        context->callback(context->buffer,count,context->userdata);
        context->z.next_out = context->buffer;
        context->z.avail_out = BUFFER;
        DEBUG_p("avail_in = %i",context->z.avail_in);
    }
    DEBUG_p("end");
}

static void stream_done(StreamInterface* sif,CompressionContext* context){
    int status;
    int count;
    DEBUG_p("begin");
    context->z.next_in = NULL;
    context->z.avail_in = 0;
    for(;;){
        DEBUG_p("before deflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
        status = sif->feed(&context->z,Z_FINISH);
        DEBUG_p("after deflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
        count = BUFFER - context->z.avail_out;
        if(!count)
            break;
        DEBUG_p("%i available to write",count);
        context->callback(context->buffer,count,context->userdata);
        context->z.next_out = context->buffer;
        context->z.avail_out = BUFFER;
    }
    DEBUG_p("end");
    sif->done(&context->z);
    free(context);
}



#if 0
CompressionContext* decompress_init(CompressionCallback callback,void* userdata){
    CompressionContext* context;
    context = (CompressionContext*)malloc(sizeof(CompressionContext));
    inflateInit(&context->z);
    context->z.next_out = context->buffer;
    context->z.avail_out = BUFFER;
    context->callback = callback;
    context->userdata = userdata;
    return context;
}

void decompress_feed(CompressionContext* context,void* data,int bytes){
    int status;
    int count;

    DEBUG_p("begin");
    context->z.next_in = data;
    context->z.avail_in = bytes;

    while(context->z.avail_in){
        for(;;){
            DEBUG_p("before inflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
            status = inflate(&context->z,Z_NO_FLUSH);
            DEBUG_p("after inflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
            count = BUFFER - context->z.avail_out;
            if(!count)
                break;
            context->callback(context->buffer,count,context->userdata);
            context->z.next_out = context->buffer;
            context->z.avail_out = BUFFER;
        }
    }
    DEBUG_p("end");
}

void decompress_done(CompressionContext* context){
    int status;
    int count;
    DEBUG_p("begin");
    for(;;){
        DEBUG_p("before inflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
        status = deflate(&context->z,Z_FINISH);
        DEBUG_p("after inflate in=%i out=%i",context->z.avail_in,context->z.avail_out);
        count = BUFFER - context->z.avail_out;
        if(!count)
            break;
        context->callback(context->buffer,count,context->userdata);
        context->z.next_out = context->buffer;
        context->z.avail_out = BUFFER;
    }
    DEBUG_p("end");
    inflateEnd(&context->z);
    free(context);
}
#endif



static StreamInterface deflate_if = {
    deflateInit_wrapper,
    deflate,
    deflateEnd
};

static StreamInterface inflate_if = {
    inflateInit_wrapper,
    inflate,
    inflateEnd
};

/* --- public functions ------------------------------------------------- */

CompressionContext* compress_init(CompressionCallback callback,void* userdata){
    return stream_init(&deflate_if,callback,userdata);
}

void compress_feed(CompressionContext* context,void* data,int bytes){
    stream_feed(&deflate_if,context,data,bytes);
}

void compress_done(CompressionContext* context){
    stream_done(&deflate_if,context);
}

CompressionContext* decompress_init(CompressionCallback callback,void* userdata){
    return stream_init(&inflate_if,callback,userdata);
}

void decompress_feed(CompressionContext* context,void* data,int bytes){
    stream_feed(&inflate_if,context,data,bytes);
}

void decompress_done(CompressionContext* context){
    stream_done(&inflate_if,context);
}
