/*
 * glade2c - convert glade user interfaces to C source
 *
 * Copyright (C) 2009 Magnus Bergman
 *
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <fcntl.h>                      /* O_RDONLY */
#include <glib.h>
#include <string.h>                     /* strcmp */
#include <unistd.h>                     /* open(), read(), close() */

#include "debug.h"
#include "glade2c.h"
#include "project.h"

/**

  Structure of a gladep file:

<glade-project>
  <name>
  <program_name>
  <gnome_support>
*/

#define BUFFER_SIZE 4096

GSList* widgetdb;

typedef struct{
    int blindlevel;
    enum {S_NONE,S_GLADEP,S_STRING,S_BOOL} state;
    union{
        char** string_p;
        xbool* bool_p;
    };
}State;

static State state;                     /**< A global state since the code doesn't need to be reentrant */



/**
  @param userdata Contains a pointer to the state, but isn't used.
*/
static void on_start_element(
  GMarkupParseContext* context,
  const char* element,
  const char** attr_name,
  const char** attr_value,
  void* userdata,
  GError** error
){
    if(state.blindlevel>0){
        ++state.blindlevel;
        DEBUG_p("increasing blind level to %i",state.blindlevel);
        return;
    }
    switch(state.state){
    case S_NONE:
        if(!strcmp(element,"glade-project")){
            state.state = S_GLADEP;
            return;
        }
        break;
    case S_GLADEP:
        if(!strcmp(element,"name")){
            state.string_p = &project.name;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"program_name")){
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"source_directory")){
            state.string_p = &project.source_directory;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"pixmaps_directory")){
            state.string_p = &project.pixmaps_directory;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"gnome_support")){
            state.bool_p = &project.gnome_support;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"gettext_support")){
            state.bool_p = &project.gettext_support;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"use_widget_name")){
            state.bool_p = &project.use_widget_name;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"output_main_file")){
            state.bool_p = &project.output_main_file;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"output_support_files")){
            state.bool_p = &project.output_support_files;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"output_build_files")){
            state.bool_p = &project.output_build_files;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"backup_source_files")){
            state.bool_p = &project.backup_source_files;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"gnome_help_support")){
            state.bool_p = &project.gnome_help_support;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"main_source_file")){
            state.string_p = &project.interface_source_file;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"main_header_file")){
            state.string_p = &project.interface_header_file;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"handler_source_file")){
            state.string_p = &project.callback_source_file;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"handler_header_file")){
            state.string_p = &project.callback_header_file;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"support_source_file")){
            state.string_p = &project.support_source_file;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"support_header_file")){
            state.string_p = &project.support_header_file;
            state.state = S_STRING;
            return;
        }else if(!strcmp(element,"output_translatable_strings")){
            state.bool_p = &project.output_translatable_strings;
            state.state = S_BOOL;
            return;
        }else if(!strcmp(element,"translatable_strings_file")){
            state.string_p = &project.translatable_strings_file;
            state.state = S_STRING;
            return;
        }
        break;
    case S_STRING:
    case S_BOOL:
        break;
    }
    DEBUG_p("unexpected tag <%s> was found",element);
    state.blindlevel = 1;
}



static void on_end_element(
  GMarkupParseContext* context,
  const char* element,
  void* userdata,
  GError** error
){
    if(state.blindlevel>0){
        --state.blindlevel;
        DEBUG_p("decreasing blind level to %i",state.blindlevel);
        return;
    }
    switch(state.state){
    case S_NONE:
        DEBUG_p("there should be no closing tags on root level");
        return;
    case S_GLADEP:
        state.state = S_NONE;
        return;
    case S_STRING:
    case S_BOOL:
        state.state = S_GLADEP;
        return;
    }
    DEBUG_p("unknown closing tag encountered");
}



static void on_text(
  GMarkupParseContext* context,
  const char* text,
  gsize text_len,  
  gpointer userdata,
  GError** error
){
    switch(state.state){
    case S_STRING:
        DEBUG_p("about to set string");
        if(!*state.string_p)
            *state.string_p = g_strndup(text,text_len);
        return;
    case S_BOOL:
        DEBUG_p("about to set bool");
        if(!state.bool_p)
            *state.bool_p = !strncmp(text,"TRUE",text_len)?false:true;
        return;
    default:
        return;
    }
}



static GMarkupParser parser = {
    on_start_element,
    on_end_element,
    on_text,
    NULL,
    NULL
};

gboolean parse_gladep(){
    GMarkupParseContext* context;
    char data[BUFFER_SIZE];
    int data_size;
    int fd;

    /* start parsning the xml-file */
    DEBUG_p("starting to parse %s",project.gladep_filename);

    fd = open(project.gladep_filename,O_RDONLY);
    if(fd<0){
        DEBUG_p("error opening %s",project.gladep_filename);
        return FALSE;
    }
    context = g_markup_parse_context_new(&parser,0,&state,NULL);
    for(;;){
        data_size = read(fd,data,BUFFER_SIZE);
        if(0==data_size){
            /* eof */
            break;
        }
        g_markup_parse_context_parse(context,data,data_size,NULL);
    }
    close(fd);
    g_markup_parse_context_free(context);

    /* read one window (top level widget) and it's children at a time */
    return TRUE;
}
