/*
 * glade2c - convert glade user interfaces to C source
 *
 * Copyright (C) 2009 Magnus Bergman
 *
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <fcntl.h>                      /* O_RDONLY */
#include <glib.h>
#include <string.h>                     /* strcmp */
#include <unistd.h>                     /* open(), read(), close() */

#include "debug.h"
#include "glade2c.h"
#include "project.h"
#include "widgetdb.h"
#include "write-source.h"

/**

  Structure of a glade file:

<glade-interface>
  <widget>
    <property>
    <child>
      <widget>
      <packing>
        <property>
    <child>
*/

#define BUFFER_SIZE 4096

typedef struct{
    int blindlevel;
    enum {S_NONE,S_GLADE,S_WIDGET,S_CHILD,S_PACKING,S_PROPERTY,S_SIGNAL} state;
    enum {PS_WIDGET,PS_PACKING} property_state;
    GNode* window_tree;
    GNode* widget_node;                 /**< The current node in the window/widget tree */
    Widget* widget;                     /**< The data of widget_node */
    Property* property;
    Signal* signal;
}State;

static State state;                     /**< A global state since the code doesn't need to be reentrant */



static void destroy_widget(Widget* widget){
    if(widget->id)
        g_free(widget->id);
    if(widget->property_list){
        /* FIXME!! destroy each property, then the propery list */
    }
    if(widget->used_property_list){
        /* FIXME!! destroy each property, then the propery list */
    }

    /* packing stuff */
    if(widget->pack_property_list){
        /* FIXME!! destroy each property, then the packing propery list */
    }
    g_free(widget);
}

static Widget* create_widget(const char** attr_name,const char** attr_value){
    GSList* iter;
    int i;
    DEBUG_p("begin");
    state.widget = g_new0(Widget,1);
    for(i=0;attr_name[i];++i){
        if(!strcmp(attr_name[i],"class")){
            state.widget->info = NULL;
            for(iter=widgetdb;iter;iter=iter->next){
                WidgetInfo* info;
                info = (WidgetInfo*)iter->data;
                if(!strcmp(attr_value[i],info->class)){
                    state.widget->info = info;
                    break;
                }
            }
            if(!state.widget->info){
                g_warning("The class %s is unknown.",attr_value[i]);
                break;
            }
        }else if(!strcmp(attr_name[i],"id")){
            state.widget->id = g_strdup(attr_value[i]);
        }
    }
    if(!state.widget->info){
        destroy_widget(state.widget);
        state.widget = NULL;
    }
    return state.widget;
}



static void create_property(const char** attr_name,const char** attr_value){
    int i;
    state.property = NULL;
    for(i=0;attr_name[i];++i){
        if(!strcmp(attr_name[i],"name")){
            WidgetInfo* widget_iter;
            /* now we have the name of the property (int attr_value[i]) */
            DEBUG_p("searching for property %s for widget %s",attr_value[i],state.widget->id);
            for(widget_iter=state.widget->info;widget_iter;widget_iter=widget_iter->parent){
                GSList* property_iter;
                /* search all defined properties for this widget class */
                DEBUG_p("widget class %s has %i defined properties",widget_iter->class,g_slist_length(widget_iter->property_list));
                for(property_iter=widget_iter->property_list;property_iter;property_iter=property_iter->next){
                    PropertyInfo* info;
                    info = (PropertyInfo*)property_iter->data;
                    if(!strcmp(attr_value[i],info->name)){
                        DEBUG_p("found property \"%s\" in class %s",info->name,widget_iter->class);
                        if(T_UNKNOWN==info->type)
                            g_warning("The property \"%s\" was found but has no type.",info->name);
                        state.property = g_new0(Property,1);
                        state.property->info = info;
                        return;
                    }
                    /* DEBUG_p("property %s of %s did not match",info->name,widget_iter->class); */
                }
            }
            g_warning("The property %s is unknown for class %s.",attr_value[i],state.widget->info->class);
            return;
        }
    }
    g_warning("The property has no name.");
}

static void create_signal(const char** attr_name,const char** attr_value){
    int i;
    state.signal = g_new0(Signal,1);
    for(i=0;attr_name[i];++i){
        if(!strcmp(attr_name[i],"name")){
            state.signal->name = g_strdup(attr_value[i]);
        }else if(!strcmp(attr_name[i],"handler")){
            state.signal->handler = g_strdup(attr_value[i]);
        }else if(!strcmp(attr_name[i],"object")){
            state.signal->object = g_strdup(attr_value[i]);
        }else if(!strcmp(attr_name[i],"after")){
            state.signal->after = strchr("TtYy",*attr_value[i])?TRUE:FALSE;
        }else if(!strcmp(attr_name[i],"lookup")){
            state.signal->lookup = strchr("TtYy",*attr_value[i])?TRUE:FALSE;
        }
    }
    if(!state.signal->name)
        g_warning("The signal has no name.");
    /* add the callback function from the signal to the callback list */
    g_hash_table_insert(project.callback_list,state.signal->handler,NULL);
}

/**
  @param userdata Contains a pointer to the state, but isn't used.
*/
static void on_start_element(
  GMarkupParseContext* context,
  const char* element,
  const char** attr_name,
  const char** attr_value,
  void* userdata,
  GError** error
){
    if(state.blindlevel>0){
        ++state.blindlevel;
        DEBUG_p("increasing blind level to %i",state.blindlevel);
        return;
    }
    switch(state.state){
    case S_NONE:
        if(!strcmp(element,"glade-interface")){
            state.state = S_GLADE;
            return;
        }
        state.blindlevel = 1;
        DEBUG_p("unexpected <%s> in root of %s",element,project.glade_filename);
        return;
    case S_GLADE:
        if(!strcmp(element,"widget")){
            create_widget(attr_name,attr_value);
            /* FIXME!! Fix create_widget so that widget is NULL rather than
               being incorrect in some way */
            if(!state.widget){
                g_warning("Unable to create widget.");
                state.blindlevel = 1;
                return;
            }else if(!state.widget->info){
                DEBUG_p("widget %s has no info",state.widget->id);
                state.blindlevel = 1;
                return;
            }
            DEBUG_p("adding widget %s (%s) as root",state.widget->id,state.widget->info->class);
            state.window_tree = state.widget_node = g_node_new(state.widget);
            state.state = S_WIDGET;
            return;
        }
        DEBUG_p("unexpected <%s> inside <glade-interface> in %s",element,project.glade_filename);
        state.blindlevel = 1;
        return;
    case S_WIDGET:
        if(!strcmp(element,"child")){
            state.widget_node = g_node_append_data(state.widget_node,state.widget = NULL);
            state.state = S_CHILD;
            return;
        }else if(!strcmp(element,"property")){
            create_property(attr_name,attr_value);
            state.state = S_PROPERTY;
            state.property_state = PS_WIDGET;
            return;
        }else if(!strcmp(element,"signal")){
            create_signal(attr_name,attr_value);
            state.state = S_SIGNAL;
            return;
        }
        DEBUG_p("unexpected <%s> inside <widget> in %s",element,project.glade_filename);
        state.blindlevel = 1;
        return;
    case S_CHILD:
        if(!strcmp(element,"widget")){
            if(state.widget_node->data){
                g_warning("More than one <widget> inside <child>, will be ignored.");
                state.blindlevel = 1;
                return;
            }
            state.widget_node->data = create_widget(attr_name,attr_value);
            /* FIXME!! Fix create_widget so that widget is NULL rather than
               being incorrect in some way */
            if(!state.widget){
                g_warning("Unable to create widget.");
                state.blindlevel = 1;
                return;
            }else if(!state.widget->info){
                DEBUG_p("widget %s has no info",state.widget->id);
                state.blindlevel = 1;
                return;
            }
            DEBUG_p("added widget %s (%s) at level %i",state.widget->id,state.widget->info->class,g_node_depth(state.widget_node));
            state.state = S_WIDGET;
            return;
        }else if(!strcmp(element,"packing")){
            if(!state.widget){
                DEBUG_p("do not add packing to unexisting widget");
                state.blindlevel = 1;
                return;
            }else if(!state.widget->info){
                DEBUG_p("do not add packing to unknown widget");
                state.blindlevel = 1;
                return;
            }
            /* perhaps add something to the widget? */
            state.state = S_PACKING;
            return;
#if 0
        }else if(!strcmp(element,"placeholder")){
            DEBUG_p("ignore placeholder");
#endif
        }
        DEBUG_p("unexpected <%s> inside <child> in %s",element,project.glade_filename);
        state.blindlevel = 1;
        return;
    case S_PACKING:
        if(!strcmp(element,"property")){
            create_property(attr_name,attr_value);
            state.state = S_PROPERTY;
            state.property_state = PS_PACKING;
            return;
        }
        DEBUG_p("unexpected <%s> inside <packing> in %s",element,project.glade_filename);
        state.blindlevel = 1;
        return;
    case S_PROPERTY:
        DEBUG_p("unexpected <%s> inside <property> in %s",element,project.glade_filename);
        state.blindlevel = 1;
        return;
    case S_SIGNAL:
        DEBUG_p("unexpected <%s> inside <signal> in %s",element,project.glade_filename);
        state.blindlevel = 1;
        return;
    }
    DEBUG_p("unexpected <%s> inside <UNKNOWN> in %s",element,project.glade_filename);
    ++state.blindlevel;
}

static void on_end_element(
  GMarkupParseContext* context,
  const char* element,
  void* userdata,
  GError** error
){
    if(state.blindlevel>0){
        --state.blindlevel;
        DEBUG_p("decreasing blind level to %i",state.blindlevel);
        return;
    }
    switch(state.state){
    case S_NONE:
        DEBUG_p("A closing tag in the root shuold not exist");
        return;
    case S_GLADE:
        DEBUG_p("</%s> should be </glade-interface>",element);
        state.state = S_NONE;
        return;
    case S_WIDGET:
        DEBUG_p("</%s> should be </widget>",element);
        /* is the widget a toplevel window? */
        if(G_NODE_IS_ROOT(state.widget_node)){
            /* Here is the right place to write the window (top level
               widget) to the source files */
            DEBUG_p("writing top level window source");
            write_window_tree(state.window_tree);
            /* FIXME!! free the whole widget_tree */
            state.state = S_GLADE;
        }else{
            state.state = S_CHILD;
        }
        return;
    case S_CHILD:
        DEBUG_p("</%s> should be </child>",element);
        DEBUG_p("widget_node = %xh",(int)state.widget_node);
#if 0
        DEBUG_p("widget = %xh",(int)state.widget);
#endif
        if(!state.widget){
            DEBUG_p("The child didn't contain any widget, which is normal");
            /* FIXME!! perhaps make sure this this node is deleted? */
#ifdef DEBUG
        }else if(G_NODE_IS_ROOT(state.widget_node)){
            DEBUG_p("this is the root node, something seems wrong");
#endif
        }else{
            /* this can (and shoul) not be done if the widget in's fully valid */
            state.widget->position = g_strdup_printf("%i",g_node_child_position(state.widget_node->parent,state.widget_node));
            state.widget->even_position = !(g_node_child_position(state.widget_node->parent,state.widget_node)&1);
            DEBUG_p("before setting PREVIOUS property");
            /* Set the "PREVIOUS" property of the previous widget. This is a
               hack to be able to make linkers which packs several widgets
               at once (like the one for the notebook). */
            if(state.widget_node->prev && state.widget_node->prev->data)
                state.widget->previous = (Widget*)state.widget_node->prev->data;
            DEBUG_p("after setting PREVIOUS property");
        }
        state.widget_node = state.widget_node->parent;
        DEBUG_p("widget_node = %xh",(int)state.widget_node);
        state.widget = (Widget*)state.widget_node->data;
        state.state = S_WIDGET;
        return;
    case S_PACKING:
        DEBUG_p("</%s> should be </packing>",element);
        state.state = S_CHILD;
        return;
    case S_PROPERTY:
        DEBUG_p("</%s> should be </property>",element);
        switch(state.property_state){
        case PS_WIDGET:
            if(state.property){
                if(!state.property->info){
                    g_warning("A property of %s has no name.",state.widget->id);
                }else if(!state.property->value){
                    g_warning("property %s of %s has no value.",state.property->info->name,state.widget->id);
                }else{
                    DEBUG_p("adding property %s=%s to %s",state.property->info->name,state.property->value,state.widget->id);
                    state.widget->property_list = g_slist_prepend(state.widget->property_list,state.property);
                }
            }
            state.state = S_WIDGET;
            break;
        case PS_PACKING:
            if(state.property){
                DEBUG_p("trying to add packing property %s=%s to %s",state.property->info->name,state.property->value,state.widget->id);
                if(!state.property->info){
                    g_warning("A packing property of %s has no name.",state.widget->id);
                }else if(!state.property->value){
                    g_warning("Packing property %s of %s has no value.",state.property->info->name,state.widget->id);
                }else{
                    DEBUG_p("adding packing property %s=%s to %s",state.property->info->name,state.property->value,state.widget->id);
                    state.widget->pack_property_list = g_slist_prepend(state.widget->pack_property_list,state.property);
                }
            }
            state.state = S_PACKING;
            break;
        }
        state.property = NULL;
        return;
    case S_SIGNAL:
        if(state.signal){
            DEBUG_p("trying to add signal %s to %s",state.signal->name,state.widget->id);
            state.widget->signal_list = g_slist_append(state.widget->signal_list,state.signal);
            state.signal = NULL;
        }
        state.state = S_WIDGET;
        return;
    default:
        DEBUG_p("don't care about </%s>",element);
        return;
    }
}

static void on_text(
  GMarkupParseContext* context,
  const char* text,
  gsize text_len,  
  gpointer userdata,
  GError** error
){
    switch(state.state){
    case S_PROPERTY:
        if(state.property)
            switch(state.property->info->type){
            case T_UNKNOWN:
                g_warning("The type of property %s is unknown",state.property->info->name);
                break;
            case T_LITERAL:
                state.property->value = g_strndup(text,text_len);
                break;
            case T_INTEGER:
            case T_REAL:
                if(text_len){
                    state.property->value = g_strndup(text,text_len);
                }else{
                    state.property->value = g_strdup("0");
                }
                break;
            case T_BOOLEAN:
                state.property->value = strchr("TtYy",*text)?"TRUE":"FALSE";
                break;
            case T_STRING:
                {
                    char format[16];
                    DEBUG_p("property type is string");
                    sprintf(format,"\"%%%is\"",text_len);
                    DEBUG_p("format is %s",format);
                    state.property->value = g_strdup_printf(format,text);
                    DEBUG_p("string is %s",state.property->value);
                }
                break;
            }
        return;
    default:
#if 0
        DEBUG_p("did not expect text in current state");
        {
            char* s;
            s = g_strndup(text,text_len);
            DEBUG_p("%s",s);
            g_free(s);
        }
#endif
        return;
    }
}

#if 0
  /* Called for strings that should be re-saved verbatim in this same
   * position, but are not otherwise interpretable.  At the moment
   * this includes comments and processing instructions.
   */
  /* text is not nul-terminated. */
  void (*passthrough)    (GMarkupParseContext *context,
                          const gchar         *passthrough_text,
                          gsize                text_len,  
                          gpointer             user_data,
                          GError             **error);

  /* Called on error, including one set by other
   * methods in the vtable. The GError should not be freed.
   */
  void (*error)          (GMarkupParseContext *context,
                          GError              *error,
                          gpointer             user_data);
#endif

static GMarkupParser parser = {
    on_start_element,
    on_end_element,
    on_text,
    NULL,
    NULL
};

/**
  Read one window (top level widget) and it's children at a time.
*/
int read_glade(){
    GMarkupParseContext* context;
    char data[BUFFER_SIZE];
    int data_size;
    int fd;

    /* start parsning the xml-file */
    DEBUG_p("starting to parse %s",project.glade_filename);

    context = g_markup_parse_context_new(&parser,0,&state,NULL);
    fd = open(project.glade_filename,O_RDONLY);
    for(;;){
        data_size = read(fd,data,BUFFER_SIZE);
        if(0==data_size){
            /* eof */
            break;
        }
        g_markup_parse_context_parse(context,data,data_size,NULL);
    }
    close(fd);
    g_markup_parse_context_free(context);

    return 0;
}
