/*
 * glade2c - convert glade user interfaces to C source
 *
 * Copyright (C) 2009 Magnus Bergman
 *
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef DEBUG
# undef DEBUG
#endif

#include <fcntl.h>                      /* O_RDONLY */
#include <glib.h>
#include <string.h>                     /* strcmp */
#include <unistd.h>                     /* open(), read(), close() */

#include "debug.h"
#include "template.h"
#include "widgetdb.h"

#ifdef USE_BUILTIN_WIDGETDB
# include "data-gtk.widgetdb.h"
#endif
#ifdef USE_COMPRESSION
# include "compression.h"
#endif

#define BUFFER 4096
#define FILENAME "data/gtk.widgetdb"

/**

  Structure of a widgetdb file:

<widgetdb>
  <widget>
    <class>
    <constructor>
    <constructor>
    <property>
      <name>
      <default>
      <setter>
*/

GSList* widgetdb;

typedef struct{
    int blindlevel;
    enum {S_NONE,S_WIDGETDB,S_WIDGET,S_CLASS,S_CONSTRUCTOR,S_LINKER,S_PARENT,S_PROPERTY,S_NAME,S_TYPE,S_DEFAULT,S_SETTER} state;
#if 0
    union{
      enum {WS_CLASS,WS_CONSTRUCTOR} widget_state;
      enum {PS_NAME,PS_DEFAULT,PS_SETTER} property_state;
    };
#endif
    WidgetInfo* widget;                 /**< The current widget */
    FunctionInfo* function;
    PropertyInfo* property;
}State;

static State state;                     /**< A global state since the code doesn't need to be reentrant */



static FunctionInfo* create_function(const char** attr_name,const char** attr_value){
    FunctionInfo* function;
    function = g_new0(FunctionInfo,1);
    for(;*attr_name;++attr_name,++attr_value){
        if(!strcmp(*attr_name,"condition")){
            function->condition = g_strdup(*attr_value);
        }
    }
    return function;
}



/**
  @param userdata Contains a pointer to the state, but isn't used.
*/
static void on_start_element(
  GMarkupParseContext* context,
  const char* element,
  const char** attr_name,
  const char** attr_value,
  void* userdata,
  GError** error
){
    if(state.blindlevel>0){
        ++state.blindlevel;
        DEBUG_p("increasing blind level to %i",state.blindlevel);
        return;
    }
    switch(state.state){
    case S_NONE:
        if(!strcmp(element,"widgetdb")){
            state.state = S_WIDGETDB;
            return;
        }
        break;
    case S_WIDGETDB:
        if(!strcmp(element,"widget")){
            /* FIXME!! try to find ancestor in tree */
            state.widget = g_new0(WidgetInfo,1);
            state.state = S_WIDGET;
            return;
        }
        break;
    case S_WIDGET:
        if(!strcmp(element,"property")){
            state.property = g_new0(PropertyInfo,1);
            state.widget->property_list = g_slist_prepend(state.widget->property_list,state.property);
            state.state = S_PROPERTY;
            return;
        }else if(!strcmp(element,"class")){
            state.state = S_CLASS;
            return;
        }else if(!strcmp(element,"constructor")){
            state.function = create_function(attr_name,attr_value);
            state.widget->constructor_list = g_slist_append(state.widget->constructor_list,state.function);
            state.state = S_CONSTRUCTOR;
            return;
        }else if(!strcmp(element,"linker")){
            state.function = create_function(attr_name,attr_value);
            state.widget->linker_list = g_slist_append(state.widget->linker_list,state.function);
            state.state = S_LINKER;
            return;
        }else if(!strcmp(element,"parent")){
            state.state = S_PARENT;
            return;
        }
        break;
    case S_CLASS:
        DEBUG_p("<class> shouldn't contains any tags");
        break;
    case S_CONSTRUCTOR:
        DEBUG_p("<constructor> shouldn't contains any tags");
        break;
    case S_LINKER:
        DEBUG_p("<linker> shouldn't contains any tags");
        break;
    case S_PARENT:
        DEBUG_p("<parent> shouldn't contain any tags");
        break;
    case S_PROPERTY:
        if(!strcmp(element,"name")){
            state.state = S_NAME;
            return;
        }else if(!strcmp(element,"type")){
            state.state = S_TYPE;
            return;
        }else if(!strcmp(element,"default")){
            state.state = S_DEFAULT;
            return;
        }else if(!strcmp(element,"setter")){
            state.state = S_SETTER;
            return;
        }
        break;
    case S_TYPE:
        DEBUG_p("<type> shouldn't contains any tags");
        break;
    case S_NAME:
        DEBUG_p("<name> shouldn't contains any tags");
        break;
    case S_DEFAULT:
        DEBUG_p("<default> shouldn't contains any tags");
        break;
    case S_SETTER:
        DEBUG_p("<setter> shouldn't contains any tags");
        break;
    }
    DEBUG_p("unexpected tag <%s> was found",element);
    state.blindlevel = 1;
}



static void on_end_element(
  GMarkupParseContext* context,
  const char* element,
  void* userdata,
  GError** error
){
    if(state.blindlevel>0){
        --state.blindlevel;
        DEBUG_p("decreasing blind level to %i",state.blindlevel);
        return;
    }
    switch(state.state){
    case S_NONE:
        DEBUG_p("there should be no closing tags on root level");
        return;
    case S_WIDGETDB:
        DEBUG_p("</%s> should be </widgetdb>",element);
        state.state = S_NONE;
        return;
    case S_WIDGET:
        DEBUG_p("</%s> should be </widget>",element);
        if(!state.widget->class){
            g_warning("Widget has no class");
            state.widget->class = g_strdup("UNKNOWN");
        }
        if(!state.widget->constructor_list){
            g_warning("Widget %s has no constructor defined",state.widget->class);
        }
        widgetdb = g_slist_prepend(widgetdb,state.widget);
        state.state = S_WIDGETDB;
        return;
    case S_CLASS:
        DEBUG_p("</%s> should be </class>",element);
        state.state = S_WIDGET;
        return;
    case S_CONSTRUCTOR:
#ifdef DEBUG
        DEBUG_p("</%s> should be </constructor>",element);
        state.function->args = template_build(state.function->string,'[',']');
        state.state = S_WIDGET;
        return;
#endif
    case S_LINKER:
        DEBUG_p("</%s> should be </linker>",element);
        state.function->args = template_build(state.function->string,'[',']');
        state.state = S_WIDGET;
        return;
    case S_PARENT:
        DEBUG_p("<%s> should be </parent>",element);
        state.state = S_WIDGET;
        return;
    case S_PROPERTY:
        DEBUG_p("</%s> should be </property>",element);
        state.state = S_WIDGET;
        return;
    case S_NAME:
        DEBUG_p("</%s> should be </name>",element);
        state.state = S_PROPERTY;
        return;
    case S_TYPE:
        DEBUG_p("</%s> should be </type>",element);
        state.state = S_PROPERTY;
        return;
    case S_DEFAULT:
        DEBUG_p("</%s> should be </default>",element);
        state.state = S_PROPERTY;
        return;
    case S_SETTER:
        DEBUG_p("</%s> should be </setter>",element);
        state.property->setter_args = template_build(state.property->setter,'[',']');
        state.state = S_PROPERTY;
        return;
    }
    DEBUG_p("unknown closing tag encountered");
}



static void on_text(
  GMarkupParseContext* context,
  const char* text,
  gsize text_len,  
  gpointer userdata,
  GError** error
){
    switch(state.state){
    case S_CLASS:
        state.widget->class = g_strndup(text,text_len);
        return;
    case S_CONSTRUCTOR:
    case S_LINKER:
        state.function->string = g_strndup(text,text_len);
        return;
    case S_PARENT:
        {
            GSList* iter;
            for(iter=widgetdb;iter;iter=iter->next){
                WidgetInfo* parent;
                parent = (WidgetInfo*)iter->data;
                if(!strncmp(text,parent->class,text_len)){
                    state.widget->parent = parent;
                    return;
                }
            }
        }
        g_warning("Unable to find parent to %s.",state.widget->class);
        return;
    case S_NAME:
        state.property->name = g_strndup(text,text_len);
        return;
    case S_TYPE:
        /* don't mind the length of text here, asume it's ok to possioble
           read a few bytes past. It will at least not affec the match. */
        if(!strcmp(text,"boolean") || !strcmp(text,"gboolean")){
            state.property->type = T_BOOLEAN;
        }else if(!strcmp(text,"literal")){
            state.property->type = T_LITERAL;
        }else if(!strcmp(text,"integer") || !strcmp(text,"gint")){
            state.property->type = T_INTEGER;
        }else if(!strcmp(text,"real") || !strcmp(text,"gfloat")){
            state.property->type = T_REAL;
        }else if(!strcmp(text,"string")){
            state.property->type = T_STRING;
        }else{
            g_warning("Unknown property type for %s of %s.",state.property->name,state.widget->class);
            state.property->type = T_UNKNOWN;
        }
        return;
    case S_DEFAULT:
        state.property->default_value = g_strndup(text,text_len);
        return;
    case S_SETTER:
        state.property->setter = g_strndup(text,text_len);
        return;
    default:
#if 0
        DEBUG_p("did not expect text in current state");
        {
            char* s;
            s = g_strndup(text,text_len);
            DEBUG_p("[%s]",s);
            g_free(s);
        }
#endif
        return;
    }
}



static GMarkupParser parser = {
    on_start_element,
    on_end_element,
    on_text,
    NULL,
    NULL
};


#if defined(USE_COMPRESSION) && defined(USE_BUILTIN_WIDGETDB)
static void parse_wrapper(void* data,int bytes,void* userdata){
    g_markup_parse_context_parse((GMarkupParseContext*)userdata,data,bytes,NULL);
}
#endif

gboolean read_widgetdb(){
    GMarkupParseContext* parse_context;

    parse_context = g_markup_parse_context_new(&parser,0,&state,NULL);
#ifdef USE_BUILTIN_WIDGETDB
# if defined(USE_COMPRESSION) && defined(data_gtk_widgetdb_is_compressed)
    {
        CompressionContext* decomp_context;
        decomp_context = decompress_init(parse_wrapper,parse_context);
        decompress_feed(decomp_context,data_gtk_widgetdb,data_gtk_widgetdb_size);
        decompress_done(decomp_context);
    }
# else
    g_markup_parse_context_parse(parse_context,data_gtk_widgetdb,data_gtk_widgetdb_size,NULL);
# endif
#else
    {
        char buffer[BUFFER];
        int bytes;
        int fd;

        /* start parsning the xml-file */
        DEBUG_p("starting to parse %s",FILENAME);
        fd = open(FILENAME,O_RDONLY);
        if(fd<0){
            DEBUG_p("error opening %s",FILENAME);
            return FALSE;
        }
        for(;;){
            bytes = read(fd,buffer,BUFFER);
            if(bytes<=0){
                /* eof or perhaps an error */
                break;
            }
            g_markup_parse_context_parse(parse_context,buffer,bytes,NULL);
        }
        close(fd);
    }
#endif
    g_markup_parse_context_free(parse_context);

    /* read one window (top level widget) and it's children at a time */
    return TRUE;
}
