/*
 * glade2c - convert glade user interfaces to C source
 *
 * Copyright (C) 2009 - 2018 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <fcntl.h>                      /* O_WRONLY */
#include <glib.h>
#include <string.h>                     /* strcmp() */
#include <unistd.h>                     /* write() */

#include "debug.h"
#include "glade2c.h"
#include "project.h"
#include "read-glade.h"

#ifdef USE_COMPRESSION
# include "compression.h"
#endif

#include "data-Makefile.am.h"
#include "data-autogen.sh.h"
#include "data-configure.ac.h"
#include "data-src-Makefile.am.h"
#include "data-stamp-h.in.h"

#include "data-src-interface.c-head.h"
#include "data-src-interface.h-head.h"
#include "data-src-interface.h-foot.h"

#define INDENT "  "
/*                                "g_object_set(" */
#define INDENT_GOBJECT_SET INDENT "             "

#define SELFVAR "self"

/* --- utility functions ------------------------------------------------ */

char* strdup_vf(const char* format,const char** str_array){
    char* s;
    int length;
    int arg;
    const char* src;
    char* dst;
    /* first calculate string length */
    length = 0;
    arg = 0;
    src = format;
    while(*src){
        if('%'==*src&&'s'==src[1]){
            length+=strlen(str_array[arg++]);
            src+=2;
        }else{
            ++length;
            ++src;
        }
    }
    s = g_new(char,length+1);
    /* then copy the string */
    arg = 0;
    src = format;
    dst = s;
    while(*src){
        if('%'==*src&&'s'==src[1]){
            length = strlen(str_array[arg]);
            memcpy(dst,str_array[arg++],length);
            src+=2;
            dst+=length;
        }else{
            *(dst++) = *(src++);
        }
    }
    *dst = '\0';
    return s;
}


/* ---------------------------------------------------------------------- */


static Widget* root_widget;



static void write_to_interface_source(const char *string){
#if 1
    dprintf(project.interface_source_fd,string);
#else
    write(project.interface_source_fd,string,strlen(string));
#endif
}

static void write_to_interface_header(const char *string){
    write(project.interface_header_fd,string,strlen(string));
}

static gboolean write_callback_function(char* callback,void* null,int fd){
    DEBUG_p("callbak: %s",callback);
    dprintf(fd,"void %s();\n",callback);
}

static gboolean write_window_variable(GNode* node,void* data){
    Widget* widget;
    char* s;
    widget = (Widget*)node->data;
    if(!widget)
        return FALSE;
    if(!widget->info)
        return FALSE;
#if 0
    s = g_strdup_printf("%s%s* %s;\n",project.global_widgets&&!project.window_widgets?"":INDENT,widget->info->class,widget->id);
#else
    s = g_strdup_printf("%sGtkWidget* %s;\n",project.global_widgets&&!project.window_widgets?"":INDENT,widget->id);
#endif
    if(project.project_mode || project.custom_interface_header){
        if(project.global_widgets || project.window_widgets){
            if(project.global_widgets)
                write_to_interface_header("extern ");
            write_to_interface_header(s);
        }
    }
    if(project.project_mode || project.custom_interface_source){
        write_to_interface_source(s);
    }
    g_free(s);
    /* FALSE means not abort the tree traversation */
    return FALSE;
}



static const char* get_property(
  const char* property_name,
  Widget* widget,
  Widget* parent,
  Widget* child,
  gboolean remove,
  gboolean use_default
){
    char* substr;
    /* DEBUG_p("property = %s",property_name); */
    substr = strchr(property_name,'.');
    if(substr){
        int len;
        len = substr-property_name;
        /* skip the dot */
        ++substr;
        if(!strncmp(property_name,"SELF",len)){
            /* this has no effect */
            return get_property(substr,widget,parent,child,remove,use_default);
        }else if(!strncmp(property_name,"PARENT",len)){
            return get_property(substr,parent,NULL,widget,remove,use_default);
        }else if(!strncmp(property_name,"CHILD",len)){
            return get_property(substr,child,widget,NULL,remove,use_default);
        }
    }

    /* first check for special names */
    if(!strcmp(property_name,"SELF")){
        return widget->id;
    }else if(!strcmp(property_name,"PARENT")){
        return parent->id;
    }else if(!strcmp(property_name,"CHILD")){
        return child->id;
    }else if(!strcmp(property_name,"POSITION")){
        return widget->position;
    }else if(!strcmp(property_name,"EVEN")){
        return widget->even_position?"TRUE":NULL;
    }else if(!strcmp(property_name,"ODD")){
        return widget->even_position?NULL:"TRUE";
    }else if(!strcmp(property_name,"PREVIOUS")){
        if(!widget->previous)
            return NULL;
        return widget->previous->id;
    }

    /* then try to find the property among the defined ones for this widget */
    {
        Property* property;
        GSList* iter;
        /* search among the ordinary properties */
        for(iter=widget->property_list;iter;iter=iter->next){
            property = (Property*)iter->data;
            if(!strcmp(property_name,property->info->name)){
                if(remove){
                    DEBUG_p("removing property \"%s\" = \"%s\"",property->info->name,property->value);
                    widget->property_list = g_slist_delete_link(widget->property_list,iter);
                    widget->used_property_list = g_slist_prepend(widget->used_property_list,property);
                }
                return property->value;
            }
        }
        /* search among the packing properties */
        for(iter=widget->pack_property_list;iter;iter=iter->next){
            property = (Property*)iter->data;
            if(!strcmp(property_name,property->info->name)){
                /* DEBUG_p("found packing property %s = %s",property_name,property->value); */
                return property->value;
            }
        }
    }
    if(!use_default)
        return NULL;
    /* loop through all widget properties (and get it's default value) */
    /* DEBUG_p("searging for \"%s\" among defaults",property_name); */
    {
        WidgetInfo* info;
        GSList* iter;
        for(info=widget->info;info;info=info->parent){
            /* DEBUG_p("searging for \"%s\" in %s",property_name,info->class); */
            for(iter=info->property_list;iter;iter=iter->next){
                PropertyInfo* property_info;
                property_info = (PropertyInfo*)iter->data;
                if(!strcmp(property_name,property_info->name) && property_info->default_value){
                    /* DEBUG_p("found property %s = %s among defaults",property_name,property_info->default_value); */
                    return property_info->default_value;
                }
            }
        }
    }
    g_warning("Unable to find property \"%s\".",property_name);
    return "NULL";
}

static FunctionInfo* get_function(const GSList* iter,Widget* widget,Widget* parent,Widget* child){
    for(;iter;iter=iter->next){
        FunctionInfo* function;
        char* start;
        char* end;
        char* s;
        const char* property;
        function = (FunctionInfo*)iter->data;
        if(!function->condition)
            return function;
        DEBUG_p("condition = %s",function->condition);
        DEBUG_p("widget = %s",widget?widget->id:"-");
        DEBUG_p("parent = %s",parent?parent->id:"-");
        DEBUG_p("child  = %s",child?child->id:"-");
        start = function->condition+1;
        end = strchr(start,']');
        if(!end)
            continue;
        s = g_strndup(start,end-start);
        property = get_property(s,widget,parent,child,FALSE,FALSE);
        DEBUG_p("var (%s) = %s",s,property?property:"[not available]");
        g_free(s);
        if(!property)
            continue;
        for(s=end+1;' '==*s;++s);
        if(!*s){
            /* the condition only requests the propert to be defined */
            DEBUG_p("the property (%s) was found",property);
            return function;
        }
        if(!strcmp(property,s)){
            DEBUG_p("the property (%s) was \"%s\"",property,s);
            return function;
        }
        DEBUG_p("the property (%s) had the wrong value",property);
    }
    return NULL;
}

static char* create_constructor(Widget* widget,Widget* parent,Widget* child){
    FunctionInfo* function;
    char* constructor;
    GSList* iter;
    char** args;
    int arg_count;
    int i;

    DEBUG_p("widget \"%s\" (%s)",widget->id,widget->info->class);
    function = get_function(widget->info->constructor_list,widget,parent,child);
    if(!function){
        g_warning("No matching constructor found.");
        return "NULL";
    }
    iter = function->args;
    arg_count = g_slist_length(iter);
    args = g_new(char*,arg_count+1);
    args[arg_count] = NULL;
    for(i=0;iter;++i,iter=iter->next){
        args[i] = get_property((char*)iter->data,widget,NULL,NULL,TRUE,TRUE);
        if(!args[i])
            args[i] = "NULL";
    }
#if 1
    constructor = strdup_vf(function->string,args);
#else
    constructor = g_strdup_vprintf(function->string,(void*)args);
#endif
    /* do not free any strings of args since none of them are copied */
    g_free(args);
    return constructor;
}

static char* create_setter(Widget* widget,Property* property){
    char* setter;
    GSList* iter;
    char** args;
    int arg_count;
    int i;

    /* DEBUG_p("begin"); */
    if(!property->info->setter)
        return NULL;

    iter = property->info->setter_args;
    arg_count = g_slist_length(iter);
    args = g_new(char*,arg_count+1);
    args[arg_count] = NULL;
    for(i=0;iter;++i,iter=iter->next){
        char* property_name;
        property_name = (char*)iter->data;
        if(!strcmp(property_name,"VALUE")){
            args[i] = property->value;
        }else{
            args[i] = get_property(property_name,widget,NULL,NULL,TRUE,FALSE);
            /* FIXME!! if args[i] is not set, there is no use to create a setter.
               But that could never happen, right? */
        }
    }
#if 1
    setter = strdup_vf(property->info->setter,args);
#else
    setter = g_strdup_vprintf(property->info->setter,args);
#endif
    /* do not free any strings of args since none of them are copied */
    g_free(args);
    return setter;
}

static char* create_linker(Widget* widget,Widget* parent,Widget* child,GNode* node){
    FunctionInfo* function;
    WidgetInfo* info;
    char* linker;
    GSList* iter;
    char** args;
    int arg_count;
    int i;

    /* DEBUG_p("begin"); */
    for(info = widget->info;;){
        function = get_function(info->linker_list,widget,parent,child);
        if(function)
            break;
        info = info->parent;
        if(!info){
            g_warning("Unable to add widget to \"%s\" (%s), it has no linker.",widget->id,widget->info->class);
            return NULL;
        }
    }

    iter = function->args;
    arg_count = g_slist_length(iter);
    /* DEBUG_p("%i args",arg_count); */
    args = g_new(char*,arg_count+1);
    args[arg_count] = NULL;
    for(i=0;iter;++i,iter=iter->next)
        args[i] = get_property((char*)iter->data,widget,NULL,child,TRUE,TRUE);
#if 1
    linker = strdup_vf(function->string,args);
#else
    linker = g_strdup_vprintf(function->string,args);
#endif
    /* do not free any strings of args since none of them are copied */
    g_free(args);
    return linker;
}

static gboolean write_window_widget(GNode* node,void* data){
    Widget* widget;
    char* constructor;

    /* copy the list of properties */
    /* some of them can be set using constructors, remove them from the list */
    /* loop throuh the rest of them */

    widget = (Widget*)node->data;
    if(!widget)
        return FALSE;
    if(!widget->info)
        return FALSE;
    g_print("\n");
    /* Write the constructor */
    DEBUG_p("creating constructor for %s",widget->id);
    constructor = create_constructor(widget,NULL,NULL);
    g_print(INDENT "%s = %s;\n",widget->id,constructor);
    /* Write the linker (put the widget inside its parent) */
    if(G_NODE_IS_ROOT(node)){
        root_widget = widget;
    }else{
        Widget* parent;
        char* linker;
        parent = (Widget*)node->parent->data;
        /* FIXME!! add to parent in some way, the packing properties might be the key */
        linker = create_linker(parent,NULL,widget,node);
        if(linker){
            g_print(INDENT "%s;\n",linker);
        }else{
            g_print(INDENT "/* add %s to %s */\n",widget->id,parent->id);
        }
    }
    /* Write all setters (set all properties */
    {
        GSList* iter;
        for(iter=widget->property_list;iter;iter=iter->next){
            Property* property;
            char* setter;
            property = (Property*)iter->data;
            setter = create_setter(widget,property);
            if(setter){
                widget->used_property_list = g_slist_prepend(widget->used_property_list,property);
                iter->data = NULL;
                g_print(INDENT "%s;\n",setter);
            }
        }
        widget->property_list = g_slist_remove_all(widget->property_list,NULL);
        /* check if there are properties not set remaining */
        if(widget->property_list){
            g_print(INDENT "g_object_set((GObject*)%s,\n",widget->id);
            for(iter=widget->property_list;iter;iter=iter->next){
                Property* property;
                property = (Property*)iter->data;
                g_print(INDENT_GOBJECT_SET "\"%s\",%s,\n",property->info->name,property->value);
            }
            g_print(INDENT_GOBJECT_SET "NULL);\n");
        }
    }
    /* Write code to link up all signals */
    {
        GSList* iter;
        if(widget->signal_list){
            for(iter=widget->signal_list;iter;iter=iter->next){
                Signal* signal;
                signal = (Signal*)iter->data;
                if(signal->after){
                    if(signal->object && signal->object[0]){
                        g_print(INDENT "g_signal_connect_data((void*)%s,\"%s\",G_CALLBACK(%s),%s,NULL,G_CONNECT_AFTER|G_CONNECT_SWAPPED);\n",
                                       widget->id,signal->name,signal->handler,signal->object);
                    }else{
                        g_print(INDENT "g_signal_connect_data((void*)%s,\"%s\",G_CALLBACK(%s),NULL,NULL,G_CONNECT_AFTER);\n",
                                       widget->id,signal->name,signal->handler);
                    }
                }else if(signal->object && signal->object[0]){
                    g_print(INDENT "g_signal_connect_swapped((void*)%s,\"%s\",G_CALLBACK(%s),%s);\n",
                                   widget->id,signal->name,signal->handler,signal->object);
                }else{
                    g_print(INDENT "g_signal_connect((void*)%s,\"%s\",G_CALLBACK(%s),NULL);\n",
                                   widget->id,signal->name,signal->handler);
                }
            }
        }
    }
    if(project.window_widgets){
        g_print(INDENT SELFVAR "->%s = %s;\n",widget->id,widget->id);
    }else if(!project.global_widgets){
        g_print(INDENT "GLADE_HOOKUP_OBJECT(%s,%s,\"%s\")",root_widget->id,widget->id,widget->id);
    }
    /* do not abort the tree traversion */
    return FALSE;
}

/**
  should only be called by the glade file parser
*/
gboolean write_window_tree(GNode* window_tree){
    char* s;

    root_widget = (Widget*)window_tree->data;

    if(!project.project_mode && !project.custom_interface_source && !project.custom_interface_header)
        return TRUE;

    if(project.project_mode || project.custom_interface_header){
        if(project.window_widgets)
            write_to_interface_header("\ntypedef struct{\n");
    }
    if(project.project_mode || project.custom_interface_source){
        g_print("\n\n\n");
    }

    /* write variables if static */
    if(project.global_widgets){
        g_node_traverse(window_tree,G_PRE_ORDER,G_TRAVERSE_ALL,-1,write_window_variable,NULL);
        g_print("\n");
    }

    /* write function header */
    if(project.window_widgets){
        s = g_strdup_printf("%s_t* create_%s()",root_widget->id,root_widget->id);
    }else{
        s = g_strdup_printf("GtkWidget* create_%s()",root_widget->id);
    }
    if(project.project_mode || project.custom_interface_source){
        g_print(s);
        g_print("{\n");
    }

    /* write variables if dynamic */
    if(!project.global_widgets)
        g_node_traverse(window_tree,G_PRE_ORDER,G_TRAVERSE_ALL,-1,write_window_variable,NULL);

    if(project.project_mode || project.custom_interface_source){
        if(project.window_widgets)
            g_print(INDENT "%s_t* " SELFVAR ";\n" INDENT SELFVAR " = g_new(%s_t,1);\n",root_widget->id,root_widget->id);
        /* write window creation stuff */
        g_node_traverse(window_tree,G_PRE_ORDER,G_TRAVERSE_ALL,-1,write_window_widget,NULL);

        if(project.window_widgets){
            g_print("\n" INDENT "return " SELFVAR ";\n}\n");
        }else{
            g_print("\n" INDENT "return %s;\n}\n",root_widget->id);
        }
    }

    if(project.project_mode || project.custom_interface_header){
        if(project.window_widgets){
            write_to_interface_header(INDENT "void* userdata;\n}");
            write_to_interface_header(root_widget->id);
            write_to_interface_header("_t;\n");
        }
        write_to_interface_header(s);
        write_to_interface_header(";\n");
    }

    g_free(s);
    return TRUE;
}

#ifdef USE_COMPRESSION
void write_wrapper(void* data,int bytes,void* userdata){
    int fd;
    fd = (int)userdata;
    DEBUG_p("about to write %i bytes",bytes);
    bytes = write(fd,data,bytes);
    DEBUG_p("%i bytes written",bytes);
}
void write_compressed(int fd,void* data,int bytes){
    CompressionContext* context;
    context = decompress_init(write_wrapper,(void*)fd);
    decompress_feed(context,data,bytes);
    decompress_done(context);
}
# define write_maybe_compressed write_compressed
#else
# define write_maybe_compressed write
#endif

/* --- the public functions --------------------------------------------- */

gboolean write_build_files(){
    static char* files[9*3] = {
      "AUTHORS",                            NULL,                NULL,
      "ChangeLog",                          NULL,                NULL,
      "Makefile.am",                        data_Makefile_am,    (char*)data_Makefile_am_size,
      "NEWS",                               NULL,                NULL,
      "README",                             NULL,                NULL,
      "autogen.sh",                         data_autogen_sh,     (char*)data_autogen_sh_size,
      "configure.ac",                       data_configure_ac,   (char*)data_configure_ac_size,
      "stamp-h.in",                         data_stamp_h_in,     (char*)data_stamp_h_in_size,
      "src" G_DIR_SEPARATOR_S "Makefile.am",data_src_Makefile_am,(char*)data_src_Makefile_am,
    };
    int fd;
    char* filename;
    int bytes;
    int i;

    if(!project.project_mode)
        return TRUE;

    /* FIXME!! put src/Makefile.am in the source direcotry instead on in /src */
    for(i=0;i<9*3;i+=3){
        filename = g_strjoin(G_DIR_SEPARATOR_S,project.basedir,files[i],NULL);
        fd = open(filename,O_WRONLY|O_CREAT|O_EXCL,0666);
        if(fd<0){
            DEBUG_p("could not create %s, propably because it exists already",filename);
        }else{
            DEBUG_p("created %s",filename);
            if(files[i+1]){
                bytes = (int)files[i+2];
#ifdef USE_COMPRESSION
                write_compressed(fd,files[i+1],bytes);
#else
                DEBUG_p("writing %i bytes to %s",bytes,filename);
                bytes = write(fd,files[i+1],bytes);
                DEBUG_p("written %i bytes to %s",bytes,filename);
#endif
            }
            close(fd);
        }
        g_free(filename);
    }
    return TRUE;
}

gboolean write_support(){
    return TRUE;
}

gboolean write_callback(){
    char* filename;
    int fd;

    if(project.project_mode || project.custom_callback_header){
        filename = g_strjoin(G_DIR_SEPARATOR_S,project.basedir,project.source_directory,project.callback_header_file,NULL);
        fd = open(filename,O_WRONLY|O_CREAT|O_TRUNC,0666);
        if(project.interface_header_fd<0){
            DEBUG_p("failed to create %s",filename);
            return FALSE;
        }
        g_free(filename);
        g_hash_table_foreach(project.callback_list,(GHFunc)write_callback_function,(void*)fd);
        close(fd);
    }

    return TRUE;
}

gboolean write_interface(){
    char* filename;

    if(project.project_mode || project.custom_interface_header){
        filename = g_strjoin(G_DIR_SEPARATOR_S,project.basedir,project.source_directory,project.interface_header_file,NULL);
        project.interface_header_fd = open(filename,O_WRONLY|O_CREAT|O_TRUNC,0666);
        if(project.interface_header_fd<0){
            DEBUG_p("failed to create %s",filename);
            return FALSE;
        }
        g_free(filename);
        write_maybe_compressed(project.interface_header_fd,data_src_interface_h_head,data_src_interface_h_head_size);
    }

    if(project.project_mode || project.custom_interface_source){
        filename = g_strjoin(G_DIR_SEPARATOR_S,project.basedir,project.source_directory,project.interface_source_file,NULL);
        project.interface_source_fd = open(filename,O_WRONLY|O_CREAT|O_TRUNC,0666);
        if(project.interface_source_fd<0){
            DEBUG_p("failed to create %s",filename);
            return FALSE;
        }
        g_free(filename);
        g_set_print_handler(write_to_interface_source);
        write_maybe_compressed(project.interface_source_fd,data_src_interface_c_head,data_src_interface_c_head_size);
        g_print("\n#include \"%s\"\n",project.callback_header_file);
        if(project.window_widgets)
            g_print("#include \"%s\"\n",project.interface_header_file);
    }

    read_glade();

    if(project.project_mode || project.custom_interface_header){
        write_maybe_compressed(project.interface_header_fd,data_src_interface_h_foot,data_src_interface_h_foot_size);
        close(project.interface_header_fd);
        DEBUG_p("done writing %s",project.interface_header_file);
    }
    if(project.project_mode || project.custom_interface_source){
        close(project.interface_source_fd);
        DEBUG_p("done writing %s",project.interface_source_file);
    }
    return TRUE;
}
